#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <glib.h>
#define __USE_XOPEN
#include <time.h>
#include <ncurses.h>
#include "microcosm.h"
#include "vectors.h"
#include "util.h"
#include "charts.h"
#include "ui_all.h"

int g_chartfd;

void decompress_coords(struct compressed_coord_t *comp, struct coord_t *c)
{
  c->x=comp->x;
  c->y=comp->y;
  c->z=comp->z;
}

int inslice(struct coord_t *min, struct coord_t *max, struct coord_t *c)
{
  return (c->x>=min->x) && (c->y>=min->y) && (c->z>=min->z) &&
    (c->x<=max->x) && (c->y<=max->y) && (c->z<=max->z);
}

int line_inslice(struct coord_t *min, struct coord_t *max,
  struct coord_t *c, int len)
{
  /*
   * Valid if: end within max-min, beg within max-min, max within end-beg
   */
  return (c->y>=min->y) && (c->z>=min->z) &&
    (c->y<=max->y) && (c->z<=max->z) &&
    ( ((c->x<=max->x) && (c->x>=min->x)) ||
    ((c->x+len<=max->x) && (c->x+len>=min->x)) || 
    ((max->x>=c->x) && (max->x<=c->x+len)) );
}

void load_chart_slice(struct coord_t *c1, struct coord_t *c2, char *buf,
  int buflen)
{
  char		magic[4]={0x00,0x00,0x00,0x00};
  gint32	compressed_coord_len;
  gint64 	bigint;
  char		chartent;
  unsigned char	chartval;
  int		calced_buflen;
  int		xoff,yoff,zoff;
  guint8	linelen;
  struct coord_t here;
  struct coord_t slicesize;
  struct compressed_coord_t cc;
  /* Sanity check buffer length against passed buflen. */
  slicesize=(struct coord_t){x: (c2->x-c1->x)+1, y: (c2->y-c1->y)+1,
    z: (c2->z-c1->z)+1};
  if ((slicesize.x<1) || (slicesize.y<1) || (slicesize.z<1)) {
    ui_error("Inverted orthehedral solid passed to load_chart_slice.");
    return;
  }
  calced_buflen=slicesize.x*slicesize.y*slicesize.z;
  if (calced_buflen>buflen) {
    ui_error("Slice bigger than slice buffer.");
    return;
  }
  xoff=0;
  while (xoff<calced_buflen) {
    buf[xoff]=0;
    xoff++;
  }
  /* Read chart. */
  lseek(g_chartfd,0,SEEK_SET);
  read(g_chartfd,&magic,4);
  if ((magic[0]=='C') && (magic[1]=='H') && (magic[2]=='R') && (magic[3]=='T')) {
    read(g_chartfd,&compressed_coord_len,sizeof(gint32));
    if (compressed_coord_len!=8) {
      ui_error("Chart compressed_coord_len != 8");
      return;
    }
    /* Skip the last-updated time. */
    read(g_chartfd,&bigint,sizeof(gint64));
    while (read(g_chartfd,&chartent,1)==1) {
      switch (chartent) {
      case CHARTENT_DOTVAL:
        read(g_chartfd,&cc,sizeof(struct compressed_coord_t));
        decompress_coords(&cc,&here);
        read(g_chartfd,&chartval,1);
        if (inslice(c1,c2,&here)) {
          xoff=here.x-c1->x;
          yoff=here.y-c1->y;
          zoff=here.z-c1->z;
          buf[xoff+(yoff*slicesize.x)+(zoff*slicesize.x*slicesize.y)]=chartval;
        }
        break;
      case CHARTENT_LINEOR:
        read(g_chartfd,&cc,sizeof(struct compressed_coord_t));
        decompress_coords(&cc,&here);
        read(g_chartfd,&chartval,1);
	read(g_chartfd,&linelen,1);
        if (line_inslice(c1,c2,&here,linelen)) {
          xoff=here.x-c1->x;
          if (xoff<0) {
            linelen+=xoff;
            xoff=0;
          }
          yoff=here.y-c1->y;
          zoff=here.z-c1->z;
          if (xoff+linelen>=slicesize.x) {
            linelen=(slicesize.x-xoff); // -1 removed
          }
          while (linelen>0) {
            buf[xoff+(yoff*slicesize.x)+(zoff*slicesize.x*slicesize.y)] |=
              chartval;
            xoff++;
            linelen--;
          }
        }
        break;
      case CHARTENT_DOTOR:
        read(g_chartfd,&cc,sizeof(struct compressed_coord_t));
        decompress_coords(&cc,&here);
        read(g_chartfd,&chartval,1);
        if (inslice(c1,c2,&here)) {
          xoff=here.x-c1->x;
          yoff=here.y-c1->y;
          zoff=here.z-c1->z;
          buf[xoff+(yoff*slicesize.x)+(zoff*slicesize.x*slicesize.y)]|=chartval;
        }
        break;
      default:
        ui_error("Unknown chart entry type.");
        return;
      }
    }
  } else {
    ui_error("Chart isn't a valid CHRT file.");
  }  
}

void write_chart_lineor(struct coord_t *start, char val, int len)
{
  struct compressed_coord_t cc;
  guint8 len8;
  if (len<=255) {
    cc=(struct compressed_coord_t){x: start->x, y: start->y, z: start->z};
    /* Line entry: CHARTENT_LINEOR, coords, val, len */
    len8=CHARTENT_LINEOR;
    write(g_chartfd,&len8,1);
    write(g_chartfd,&cc,sizeof(struct compressed_coord_t));
    write(g_chartfd,&val,1);
    len8=len;
    write(g_chartfd,&len8,1);
  } else {
    ui_error("Draw within the lines.");
  }
}

void write_chart_dot(struct coord_t *start, char val)
{
  struct compressed_coord_t cc=
    (struct compressed_coord_t){x: start->x, y: start->y, z: start->z};
  char c=CHARTENT_DOTVAL;
  write(g_chartfd,&c,1);
  write(g_chartfd,&cc,sizeof(struct compressed_coord_t));
  write(g_chartfd,&val,1);
}

void write_chart_dotor(struct coord_t *start, char val)
{
  struct compressed_coord_t cc=
    (struct compressed_coord_t){x: start->x, y: start->y, z: start->z};
  char c=CHARTENT_DOTOR;
  write(g_chartfd,&c,1);
  write(g_chartfd,&cc,sizeof(struct compressed_coord_t));
  write(g_chartfd,&val,1);
}

void append_chart_dotor(struct coord_t *start, char val)
{
  lseek(g_chartfd,0,SEEK_END);
  write_chart_dotor(start,val);
}

void write_chart_line(struct coord_t *start, char val, int len)
{
  if (len>1) {
    write_chart_lineor(start, val, len);
  } else if (len==1) {
    write_chart_dot(start, val);
  } else {
    ui_error("Please to keep the universe right-side-up.");
  }
}

void write_chart_slice(char *slice, struct coord_t *c,
  struct coord_t *slicesize)
{
  struct coord_t d;
  struct coord_t linestart=*c;
  char currline=CHART_BLANK;
  int slicepos;
  lseek(g_chartfd,0,SEEK_END);
  d.z=0;
  while (d.z<slicesize->z) {
    d.y=0;
    while (d.y<slicesize->y) {
      d.x=0;
      linestart=(struct coord_t){x: c->x, y: c->y+d.y, z: c->z+d.z};
      while (d.x<slicesize->x) {
        slicepos=d.x+(d.y*slicesize->x)+(d.z*slicesize->y*slicesize->x);
        if (slice[slicepos]!=currline) {
          if (currline!=CHART_BLANK) {
            write_chart_line(&linestart,currline,(d.x+c->x)-linestart.x);
          }
          currline=slice[slicepos];
          linestart=(struct coord_t){x: c->x+d.x, y: c->y+d.y, z: c->z+d.z};
        }
        d.x++;
      }
      if ((currline!=CHART_BLANK) && (linestart.x<d.x+c->x)) {
        write_chart_line(&linestart,currline,d.x+c->x-linestart.x);
      }
      currline=CHART_BLANK;
      d.y++;
    }
    d.z++;
  }
}

void open_chart_file()
{
  GString *file=g_string_new(g_Config.datapath->str);
  struct stat st;
  char c[4]={'C','H','R','T'};
  gint32 bitlen;
  gint64 chartstarted;
  g_string_append(file,MAKROKOSMOS_CHARTS_PATH);
  g_string_append(file,g_Player.name->str);
  g_chartfd=open(file->str, O_RDWR|O_CREAT, g_Config.datamode);
  if (g_chartfd<0) {
    ui_error("Warning: Unable to open chart.");
  } else {
    fchmod(g_chartfd,g_Config.datamode);
    if (fstat(g_chartfd,&st)==0) {
      if (st.st_size==0) {
        /*
	 * Header is: Text CHRT, int32 length in bits of one dimension,
	 * int64 last-updated time in machine order.
         */
        write(g_chartfd,&c,4);
        bitlen=8;
        write(g_chartfd,&bitlen,sizeof(gint32));
        chartstarted=time(NULL);
        write(g_chartfd,&chartstarted,sizeof(gint64));
      }
    } else {
      ui_error("fstat of chart file failed.");
    }
  }
  g_string_free(file,TRUE);
}

void close_chart_file()
{
  close(g_chartfd);
}

void add_log_to_chart(struct gstrlist_t *list, time_t *lastupdated)
{
  int		x=0;
  struct tm	tm;
  time_t	then;
  GString	*cmd;
  GString	*sect;
  struct coord_t c;
  lseek(g_chartfd,0,SEEK_END);
  while (x<list->num) {
    strptime(&list->strings[x]->str[0],"%a %b %d %T %Y",&tm);
    then=mktime(&tm);
    if (then>*lastupdated) {
      split(&list->strings[x]->str[26],&cmd,&sect);
      if (strcmp(cmd->str,"System generated")==0) {
        string_to_coords(sect,&c);
        write_chart_dotor(&c,CHART_BODIES);
      } else {
        ui_error_sss("Unhandled log entry \"",&list->strings[x]->str[26],
          "\"");
      }
      g_string_free(cmd,TRUE);
      g_string_free(sect,TRUE);
      *lastupdated=then;
    }
    x++;
  }
}
