/*
 *
 *  Makrokosmos sector file.  (C) 2000 Deekoo L., released under GPL.
 *
 *  Should contain all lowlevel routines pertaining to sectors,
 *  bodies, and zones.
 *
 */

#include <sys/file.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <ncurses.h>
#include <string.h>
#include "microcosm.h"
#include "vectors.h"
#include "sectors.h"
#include "util.h"
#include "ui_all.h"

int g_sector_errno=0;

struct chem_prob_t {
  char  *str;
  int   mintemp;
  int   maxtemp;
  float odds;
};

/* Gaseous.  Numbers are guesses. */
struct chem_prob_t g_ChemProbabilities[] = {
        {"H2",3,1000,.4}, {"He",-1,100000,.1}, 
	{"Li2O",80,190,.01}, {"Be",40,340,.01}, {"B",50,400,.01}, 
	{"CH4",40,350,.2}, {"C8H18",280,390,.01}, 
        {"CO",15,160,.07}, {"CO2",270,1400,.5}, /* C: 6 */
        {"N2",250,800,.4}, {"O2",240,470,.09}, {"F",20,300,.01},
        {"Ne",20,300,.01}, {"NaCl",375,420,.006}, 
        {"SiO2",1300,1600,.1}, {"Cl",350,420,.01}, {"Ar",40,350,.05},
        {"Fe",3135,2000000,1},
        {"Kr",70,400,.01}, 
        {"Xe",80,440,.02},
	{"Bi",1837,1200000,.01},
	{"Th",5063,1200000,.005},
	{"Uup",3540,12000000,.00001},
	{NULL,0,0,0} };

/* Liquids */

struct chem_prob_t g_LiqudProbabilities[] = {
/* 90 */	{"Th"/*??*/,3100,4790+273,.005},
		{"ThO2"/*??*/,3573 /*known*/,7000/*guessed*/,.005}
	};

double extract_amt(const char *z, const char *chem)
{
  char *z2=strstr(z,chem);
  if (z2) {
    while ((z2[0]!='\0') && (z2[0]!='=')) {
      z2++;
    }
    if (z2[0]=='=') {
      z2++;
    }
    return atof(z2);
  } else {
    return 0;
  }
}

double spherical_area(double radius)
{
  return (pow(radius,2)*M_PI*4);
}

double sphere_insolation(struct body_t *b, float radius)
{
  double luminosity;
  if (radius<1) radius=1;
  /* In L[sun] and M[sun] */
  luminosity=pow(b->mass/1.99E30,3.5);
  luminosity*=4E26; /* (Watts[Sun]) */
  luminosity/=spherical_area(radius*1000);
  /* Watts/M2: 1414/258K */
  /* Equiv to insolation at (orbradius/sqrt(luminosity)) from Sun */
  /* Blackbody of 1L[sun] @ 1.5E8(Orb[Earth]): 255K */
  return luminosity/5;
}

double external_insolation(struct sector_t *sect, int body)
{
  int currbody=body;
  int lastbody;
  double retval=0;
  while (sect->bodies[currbody].primary>=0) {
    lastbody=currbody;
    currbody=sect->bodies[currbody].primary;
    retval+=sphere_insolation(&sect->bodies[currbody],
      sect->bodies[lastbody].orbradius);
  }
  return retval;
}

double body_temperature(struct sector_t *sect, int body)
{
  return sphere_insolation(&sect->bodies[body],sect->bodies[body].radius)+
    external_insolation(sect,body);
}

int open_sector(char *coords)
{
  int	  fd;
  GString *sectorfile=g_string_new(g_Config.datapath->str);
  g_string_append(sectorfile,MAKROKOSMOS_SECTOR_PATH);
  g_string_append(sectorfile,coords);
  fd=open(sectorfile->str,O_RDWR);
  g_string_free(sectorfile,TRUE);
  if (fd>=0) {
    if (flock(fd,LOCK_EX)!=0) {
      g_sector_errno=errno;
      close(fd);
      return -1;
    }
  } else {
    g_sector_errno=errno;
    switch (g_sector_errno) {
    case ENOENT:
    case ENOTDIR:
      fd=create_sector(coords);
      break;
    default:
      ui_config_error("Error in open_sector()",g_sector_errno);
    }
  }
  return fd;
}

int create_sector(char *coords)
{
  int		fd;
  int		x;
  GString	*sectorfile=g_string_new(g_Config.datapath->str);
  g_string_append(sectorfile,MAKROKOSMOS_SECTOR_PATH);
  x=sectorfile->len+1;
  g_string_append(sectorfile,coords);
  while (x<sectorfile->len) {
    if (sectorfile->str[x]==G_DIR_SEPARATOR) {
      sectorfile->str[x]='\0';
      mkdir(sectorfile->str,g_Config.dirmode);
      chmod(sectorfile->str,g_Config.dirmode);
      sectorfile->str[x]=G_DIR_SEPARATOR;
    }
    x++;
  }
  fd=open(sectorfile->str,O_RDWR|O_CREAT);
  g_string_free(sectorfile,TRUE);
  if (fd>=0) {
    fchmod(fd,g_Config.datamode);
    if (flock(fd,LOCK_EX)!=0) {
      close(fd);
      return -1;
    }
  } else {
    printw("Error in create_sector()\n");
    refresh();
  }
  return fd;
}

void add_zone(struct body_t *b, char *z)
{
  b->numzones++;
  b->zones=g_realloc(b->zones,b->numzones*sizeof(GString *));
  b->zones[b->numzones-1]=g_string_new(z);
}

void extend_body_chain(struct sector_t *sect, int numbodies)
{
  if (numbodies>=sect->numbodies) {
    sect->bodies = g_realloc(sect->bodies,sizeof(struct body_t)*(numbodies+1));
    while (sect->numbodies<=numbodies) {
      sect->bodies[sect->numbodies]=null_body;
      sect->numbodies++;
    }
  }
}

void add_orbital_fleet(struct body_t *body, char *fleet)
{
  body->numofleets++;
  body->ofleets=g_realloc(body->ofleets,body->numofleets*sizeof(GString *));
  body->ofleets[body->numofleets-1]=g_string_new(fleet);
}

void add_landed_fleet(struct body_t *body, char *fleet)
{
  body->numlfleets++;
  body->lfleets=g_realloc(body->lfleets,body->numlfleets*sizeof(GString *));
  body->lfleets[body->numlfleets-1]=g_string_new(fleet);
}

void add_colony(struct body_t *body, const char *colony)
{
  body->numcolonies++;
  body->colonies=g_realloc(body->colonies,body->numcolonies*sizeof(GString *));
  body->colonies[body->numcolonies-1]=g_string_new(colony);
}

void add_fleet(struct sector_t *sect, char *fleet)
{
  sect->numfleets++;
  sect->fleets=g_realloc(sect->fleets,sect->numfleets*sizeof(GString *));
  sect->fleets[sect->numfleets-1]=g_string_new(fleet);
}

void read_sector(int fd, struct sector_t *sect)
{
  FILE *f;
  int fd2;
  int bodynum=-1;
  GString *cmd;
  GString *args;
  char	buf[1024];
  int	len;
  *sect=null_sector;
  lseek(fd,0,SEEK_SET);
  fd2=dup(fd);
  if (fd2<0) {
    printw("read_sector(){... dup(); ...} failed (errno %d).\n",errno);
    refresh();
    return;
  }
  f=fdopen(fd2,"r");
  if (f) {
    while (fgets(&buf[0],1024,f)) {
      len=strlen(&buf[0]);
      if ((len) && (buf[len-1]=='\n')) {
        buf[len-1]='\0';
      }
      if (buf[0]!='#') {
        split(&buf[0],&cmd,&args);
        if (strcmp(cmd->str,"Apogee")==0) {
          sect->bodies[bodynum].apogee=atof(args->str);
        } else if (strcmp(cmd->str,"Body")==0) {
          bodynum=atoi(args->str);
          if (bodynum>=sect->numbodies) {
            extend_body_chain(sect,bodynum);
          }
        } else if (strcmp(cmd->str,"Class")==0) {
          sect->bodies[bodynum].class=atoi(args->str);
        } else if (strcmp(cmd->str,"Colony")==0) {
          add_colony(&sect->bodies[bodynum],args->str);
        } else if (strcmp(cmd->str,"Day")==0) {
          sect->bodies[bodynum].day=atof(args->str);
        } else if (strcmp(cmd->str,"Desc")==0) {
          g_string_assign_interpolated(sect->bodies[bodynum].desc,args->str);
        } else if (strcmp(cmd->str,"EqInc")==0) {
          sect->bodies[bodynum].eqinc=atof(args->str);
        } else if (strcmp(cmd->str,"Fleet")==0) {
          add_fleet(sect,args->str);
        } else if (strcmp(cmd->str,"G")==0) {
          sect->bodies[bodynum].grav=atof(args->str);
	} else if (strcmp(cmd->str,"LandFleet")==0) {
          add_landed_fleet(&sect->bodies[bodynum],args->str);
        } else if (strcmp(cmd->str,"Mass")==0) {
          sect->bodies[bodynum].mass=atof(args->str);
        } else if (strcmp(cmd->str,"MaxRadius")==0) {
          sect->bodies[bodynum].maxradius=atoi(args->str);
        } else if (strcmp(cmd->str,"MaxTemp")==0) {
          sect->bodies[bodynum].maxtemp=atoi(args->str);
        } else if (strcmp(cmd->str,"MinRadius")==0) {
          sect->bodies[bodynum].minradius=atoi(args->str);
        } else if (strcmp(cmd->str,"MinTemp")==0) {
          sect->bodies[bodynum].mintemp=atoi(args->str);
        } else if (strcmp(cmd->str,"Name")==0) {
          g_string_assign(sect->bodies[bodynum].name,args->str);
        } else if (strcmp(cmd->str,"OrbEcc")==0) {
          sect->bodies[bodynum].orbecc=atof(args->str);
        } else if (strcmp(cmd->str,"OrbInc")==0) {
          sect->bodies[bodynum].orbinc=atof(args->str);
	} else if (strcmp(cmd->str,"OrbFleet")==0) {
          add_orbital_fleet(&sect->bodies[bodynum],args->str);
        } else if (strcmp(cmd->str,"OrbRadius")==0) {
          sect->bodies[bodynum].orbradius=atof(args->str);
        } else if (strcmp(cmd->str,"Owner")==0) {
          g_string_assign(sect->bodies[bodynum].owner,args->str);
        } else if (strcmp(cmd->str,"Perigee")==0) {
          sect->bodies[bodynum].perigee=atof(args->str);
        } else if (strcmp(cmd->str,"Primary")==0) {
          sect->bodies[bodynum].primary=atoi(args->str);
        } else if (strcmp(cmd->str,"Radius")==0) {
          sect->bodies[bodynum].radius=atoi(args->str);
        } else if (strcmp(cmd->str,"Retrograde")==0) {
          sect->bodies[bodynum].retrograde=1;
        } else if (strcmp(cmd->str,"SectorName")==0) {
          g_string_assign(sect->name,args->str);
        } else if (strcmp(cmd->str,"SectorOwner")==0) {
          g_string_assign(sect->owner,args->str);
        } else if (strcmp(cmd->str,"Year")==0) {
          sect->bodies[bodynum].year=atof(args->str);
        } else if (strcmp(cmd->str,"Zone")==0) {
          add_zone(&sect->bodies[bodynum],args->str);
        } else {
          if (strcmp(cmd->str,"")) {
            ui_config_error_sss("Invalid command '",&buf[0],"' in sector file.",0);
            shutdown_game();
          }
        }
        g_string_free(cmd,TRUE);
        g_string_free(args,TRUE);
      }
    }
    lseek(fd2,0,SEEK_SET);
    fclose(f);
  }
}

void write_body(FILE *f, struct body_t *b)
{
  int x;
  x=0;
  while (x<b->numcolonies) {
    fprintf(f,"Colony=%s\n",b->colonies[x++]->str);
  }
  x=0;
  while (x<b->numlfleets) {
    fprintf(f,"LandFleet=%s\n",b->lfleets[x++]->str);
  }
  if (b->class) {
    fprintf(f,"Class=%d\n",b->class);
  }
  if (b->desc->len) {
    fprintf(f,"Desc=");
    write_string_interpolated(f,b->desc->str);
    fprintf(f,"\n");
  }
  if (b->mass) {
    fprintf(f,"Mass=%e\n",b->mass);
  }
  if (b->maxtemp) {
    fprintf(f,"MaxTemp=%e\n",b->maxtemp);
  }
  if (b->mintemp) {
    fprintf(f,"MinTemp=%e\n",b->mintemp);
  }
  if (b->name->len) {
    fprintf(f,"Name=%s\n",b->name->str);
  }
  if (b->orbecc) {
    fprintf(f,"OrbEcc=%f\n",b->orbinc);
  }
  if (b->orbinc) {
    fprintf(f,"OrbInc=%f\n",b->orbinc);
  }
  if (b->day) {
    fprintf(f,"Day=%f\n",b->day);
  }
  if (b->year) {
    fprintf(f,"Year=%f\n",b->year);
  }
  if (b->eqinc) {
    fprintf(f,"EqInc=%f\n",b->year);
  }
  x=0;
  while (x<b->numofleets) {
    fprintf(f,"OrbFleet=%s\n",b->ofleets[x++]->str);
  }
  if (b->owner->len) {
    fprintf(f,"Owner=%s\n",b->owner->str);
  }
  if (b->primary>=0) {
    fprintf(f,"Primary=%d\n",b->primary);
  }
  if (b->grav) {
    fprintf(f,"G=%e\n",b->grav);
  }
  if (b->orbradius) {
    fprintf(f,"OrbRadius=%f\n",b->orbradius);
  }
  if (b->minradius) {
    fprintf(f,"MinRadius=%d\n",b->minradius);
  }
  if (b->radius) {
    fprintf(f,"Radius=%d\n",b->radius);
  }
  if (b->maxradius) {
    fprintf(f,"MaxRadius=%d\n",b->maxradius);
  }
  if (b->apogee) {
    fprintf(f,"Apogee=%f\n",b->apogee);
  }
  if (b->perigee) {
    fprintf(f,"Perigee=%f\n",b->perigee);
  }
  if (b->retrograde) {
    fprintf(f,"Retrograde\n");
  }
  if (b->zones) {
    x=0;
    while (x<b->numzones) {
      fprintf(f,"Zone=%s\n",b->zones[x]->str);
      x++;
    }
  }
}

void write_sector(int fd, struct sector_t *sect)
{
  FILE *f;
  int x=0;
  int fd2;
  ftruncate(fd,0);
  fd2=dup(fd);
  if (fd2<0) {
    printw("oops.\n");
    refresh();
    return;
  }
  f=fdopen(fd2,"w");
  if (f) {
    if (sect->name->len) {
      fprintf(f,"SectorName=%s\n",sect->name->str);
    }
    if (sect->owner->len) {
      fprintf(f,"SectorOwner=%s\n",sect->owner->str);
    }
    if (sect->comments->len) {
      fprintf(f,"%s",sect->comments->str);
    }
    while (x<sect->numbodies) {
      fprintf(f,"Body=%d\n",x);
      write_body(f,&sect->bodies[x++]);
    }
    x=0;
    while (x<sect->numfleets) {
      fprintf(f,"Fleet=%s\n",sect->fleets[x++]->str);
    }
    fclose(f);
  } else {
    printw("Error %d (%s) saving sector to fd %d.\n",errno,
      strerror(errno),fd);
    refresh();
  }  
}

void close_sector(int fd)
{
  flock(fd,LOCK_UN);
  close(fd);
}

void load_sector(char *coords, struct sector_t *sect)
{
  int	fd;
  fd=open_sector(coords);
  if (fd>=0) {
    read_sector(fd,sect);
    close_sector(fd);
  } else {
    if (g_sector_errno==ENOENT) {
      fd=create_sector(coords);
      if (fd>=0) {
        read_sector(fd,sect);
        close_sector(fd);
      } else {
        printw("Error creating sector %s\n",coords);
        refresh();
        exit(g_sector_errno);
      }
    } else {
      printw("Error opening sector %s\n",coords);
      refresh();
      exit(errno);
    }
  }
}

void dispose_body(struct body_t *b)
{
  int x=0;
  g_string_free(b->name,TRUE);
  g_string_free(b->owner,TRUE);
  g_string_free(b->desc,TRUE);
  while (x<b->numofleets) {
    g_string_free(b->ofleets[x],TRUE);
    x++;
  }
  x=0;
  while (x<b->numlfleets) {
    g_string_free(b->lfleets[x],TRUE);
    x++;
  }
  x=0;
  while (x<b->numcolonies) {
    g_string_free(b->colonies[x],TRUE);
    x++;
  }
  x=0;
  while (x<b->numzones) {
    g_string_free(b->zones[x],TRUE);
    x++;
  }
}

void dispose_sector(struct sector_t *sect)
{
  int x=0;
  g_string_free(sect->name,TRUE);
  g_string_free(sect->owner,TRUE);
  g_string_free(sect->comments,TRUE);
  while (x<sect->numbodies) {
    dispose_body(&sect->bodies[x]);
    x++;
  }
  g_free(sect->bodies);
  x=0;
  while (x<sect->numfleets) {
    g_string_free(sect->fleets[x],TRUE);
    x++;
  }
  g_free(sect->fleets);
}

int zone_chem_index(struct zone_t *z, char *chem)
{
  int x=0;
  while (x<z->numchems) {
    if (strcmp(z->chems[x]->str,chem)==0) {
      return x;
    }
    x++;
  }
  return -1;
}

void append_chem_amt(struct zone_t *z, char *chem, int amt)
{
  z->numchems++;
  z->chems=g_realloc(z->chems,z->numchems*sizeof(GString *));
  z->chems[z->numchems-1]=g_string_new(chem);
  z->amts=g_realloc(z->amts,z->numchems*sizeof(int));
  z->amts[z->numchems-1]=amt;
}

void add_chem_amt(struct zone_t *z, char *chem, int amt)
{
  int x=zone_chem_index(z,chem);
  if (x==-1) {
    append_chem_amt(z,chem,amt);
  } else {
    z->amts[x]+=amt;
  }
}

void zone_to_str(struct zone_t *z, GString *zonestr)
{
  int x=0;
  GString *s=g_string_new("");
  if (z->mintemp<0) {
    ui_error("Zone minimum temp has snuck below 0K.  Fixing.");
    z->mintemp=0;
  }
  g_string_sprintf(zonestr,"t-=%e,t+=%e,p-=%f,p+=%f,",
    z->mintemp,z->maxtemp,z->minpress,z->maxpress);
  while (x<z->numchems) {
    g_string_sprintf(s,"%s=%d,",z->chems[x]->str,z->amts[x]);
    g_string_append(zonestr,s->str);
    x++;
  }
  g_string_free(s,TRUE);
}

void dispose_zone(struct zone_t *z)
{
  int x=0;
  while (x<z->numchems) {
    g_string_free(z->chems[x],TRUE);
    x++;
  }
  g_free(z->chems);
  g_free(z->amts);
}

void generate_zone(GString *zonestr, double mintemp, double maxtemp, 
	double minpress, double maxpress)
{
  struct zone_t z=null_zone;
  int		x=0;
  int		amttoalloc=10000;
  int		amt;
  int		loops=5;
  z.mintemp=mintemp;
  z.maxtemp=maxtemp;
  z.minpress=minpress;
  z.maxpress=maxpress;
  while (amttoalloc>0) {
    if (g_ChemProbabilities[x].str) {
      if ((maxtemp>g_ChemProbabilities[x].mintemp) &&
        (mintemp<g_ChemProbabilities[x].maxtemp) &&
        (probability(g_ChemProbabilities[x].odds))) {
        amt=lesser(randval_i(10000),amttoalloc);
        add_chem_amt(&z,g_ChemProbabilities[x].str,amt);
        amttoalloc-=amt;
      }
      x++;
    } else {
      if (--loops<=0) {
        amttoalloc=0;
      }
      x=0;
    }
  }
  zone_to_str(&z,zonestr);
  dispose_zone(&z);
}

void fix_sector(struct sector_t *s)
{
  int	x=0;
  int	y;
  while (x<s->numbodies) {
    if (s->bodies[x].primary>=s->numbodies) {
      /* Move any body whose primary isn't in the sector to
		loose-in-the-sector */
      s->bodies[x].primary=-1;
    }
    if (s->bodies[x].orbradius<=0) {
      s->bodies[x].orbradius=1;
      y=0;
      if (s->bodies[x].primary>=0) {
        s->bodies[x].orbradius+=s->bodies[s->bodies[x].primary].radius;
      }
      while (y<x) {
        if (s->bodies[x].primary==s->bodies[y].primary) {
          s->bodies[x].orbradius+=s->bodies[y].orbradius;
        }
        y++;
      }
    }
    x++;
  }
}

void generate_zones(char *coords, int body)
{
  int			fd;
  struct sector_t	sect;
  double		basetemp;
  double		lowtemp,hightemp;
  float			retention;
  fd=open_sector(coords);
  if (fd>=0) {
    read_sector(fd,&sect);
    if (sect.bodies[body].zones==NULL) {
      /* Determine ambient conditions */
      /* For now, generate_zones generates one zone.  Later, it should
	make more on larger worlds.. */
      retention=1;
      if (probability(0.1)) {
        /* Albedo, as supplied by one's arse. */
        retention-=((128-randval_i(256))/1024);
      }
      basetemp=body_temperature(&sect,body);
      /* And greenhouse effect, from the same source. */
      if (probability(0.2)) {
        retention+=(cbrt(sect.bodies[body].grav)/2)/(basetemp/300);
      }
      if (basetemp<0) {
        ui_error("Negative insolation in sector, probably caused by incomplete sector data.  Fixing sector.");
        fix_sector(&sect);
        basetemp=greater(randval_i(40),body_temperature(&sect,body));
      }
      retention+=cbrt(sect.bodies[body].grav)/10;
      hightemp=(basetemp*retention)+sect.bodies[body].maxtemp;
      lowtemp=lesser_d(basetemp,basetemp*retention)+sect.bodies[body].mintemp;
      add_zone(&sect.bodies[body],"");
      generate_zone(sect.bodies[body].zones[0],lowtemp,hightemp,
        cbrt(sect.bodies[body].grav)/2,cbrt(sect.bodies[body].grav));
    }
    write_sector(fd,&sect);
    close_sector(fd);
    dispose_sector(&sect);
  }
}

int fleet_in_sector(struct fleet_t *fleet, struct sector_t *sect)
{
  if (fleet->body>=sect->numbodies) {
    return 0;
  }
  if (fleet->zone>=sect->bodies[fleet->body].numzones) {
    return 0;
  }
  return 1;
}

/*
void generate_satellite(struct sector_t *sect, int primary, int nesting)
{
  float orb;
  float mass=sect->bodies[primary].mass/(10+randval_i(100)+randval_i(1000));
  int	snum=sect->numbodies;
  extend_body_chain(sect,snum);
  sect->bodies[snum].mass=mass;
  sect->bodies[snum].primary=primary;
  if (snum>0 && sect->bodies[snum-1].primary==primary) {
    orb=sect->bodies[primary].radius+sect->bodies[snum-1].orbradius+
      ((float)sect->bodies[snum-1].radius*1.2);
    printw("PrelimOrbit (prevsat found): %f\n",orb);
  } else {
    orb=pow(sect->bodies[primary].mass/1.99E30,4);
--     3.5 would be luminosity.  Tesseracted min. --
    printw("PrelimOrbit (no prevsat): %f\n",orb);
  }
  orb+=randval_d(orb*40);
  printw("Final orbit: %f\n",orb);
  sect->bodies[snum].orbradius=orb;
  if (probability(0.7-(0.2*nesting))) {
    generate_satellite(sect,snum,nesting+1);
  }
}*/

void initialize_body_class(struct body_t *b)
{
  if (b->mass>1.6E29) {
    /* Stars */
    if ((b->mass>1E31) && probability(0.1)) {
      if (b->mass>4E31) {
        b->class=BCLASS_BLACKHOLE;
        // Add code to calculate black hole radius. */
      } else {
        b->mass/=6.6;
        if (b->mass<6E30) {
          if (b->mass<2.8E30) {
            b->class=BCLASS_WHITEDWARF;
          } else {
            b->class=BCLASS_NEUTRONSTAR;
          }
        } else {
          b->class=BCLASS_WORMHOLE;
        }
      }
    } else {
      b->class=BCLASS_STAR;
    }
  } else {
    if (b->mass>4.75E28) {
      b->class=BCLASS_BROWNDWARF;
      /* Add actual brown dwarf mass req based on deuterium fusion
       * criterion. 
       */
    } else {
      /* Planets and the like */
      if (b->mass>1e26) {
        b->class=BCLASS_GASGIANT;
      } else if (b->mass<1e25) {
        b->class=BCLASS_TERRESTRIAL;
      } else {
        if (probability((1e-26)*(1e26-b->mass))) {
          b->class=BCLASS_GASGIANT;
        } else {
          b->class=BCLASS_TERRESTRIAL;
        }
      }
    }
  }
}

void initialize_body(struct body_t *b)
{
  /*  Masses from here.  All in Kg.
   *  http://www.astro.psu.edu/users/niel/scales/scales1.ascii
   * 1.6E29: Minimum mass for hydrogen fusion.
   * 1.99E30: Solar mass
   * 2.8E30: Max white dwarf mass
   * 6.0E30: Max neutron star
   * 4E31: Mass for star which becomes black hole.
   * 1.0E32: Plaskett's Star, largest.
   */
  if (b->mass<=0) {
    ui_error("Body with mass of 0 or lower passed to body initialization code.");
    return;
  }
  if (b->class==0) {
    initialize_body_class(b);
  }
  if (b->radius<=0) {
    switch (b->class) {
    case BCLASS_STAR:
      /* This has been claimed for main seq stars.  Doesn't cover others.
       * http://super.colorado.edu/~astr1020/l5body.html
       * I probably mangled the equation. R/R[Sun] = (M/M[Sun])^0.8)].
       */
      b->radius = 696000*(pow(b->mass/1.99e30,0.8));
      break;
    case BCLASS_BROWNDWARF:
      b->radius=randval_i(40000)+50000;
      // Need actual brown dwarf radius equation.
      break;
    case BCLASS_GASGIANT:
      if (b->mass>2.5e27) {
        b->radius = ((70000/(b->mass/(1.9e27/65000)))*70000)+randval_i(10000);
      } else {
        b->radius = b->mass/(1.9e27/70000);
      }
      break;
    case BCLASS_TERRESTRIAL:
      b->radius = b->mass/(5.9e24/(6000+randval_i(2000)));
      if (b->mass>9e23) b->radius+=randval_i(100);
      if (b->mass>9e24) b->radius+=randval_i(1000);
      break;
    case BCLASS_WHITEDWARF:
      b->radius = (b->mass/6000)/(1.99E30/6000);
	/* M(.)==R(+), M/6000/(M(.)/6000), 1/pow(b->mass,.333); */
      break;
    }
  }
}

void boil_orbiters(struct sector_t *sect, int sat)
{
  float temp=external_insolation(sect,sat);
  int oldclass;
  switch (sect->bodies[sat].class) {
  case BCLASS_STAR:
    while (temp>2e6) {
      sect->bodies[sat].mass-=(sect->bodies[sat].mass/32);
      sect->bodies[sat].orbradius*=2;
      temp=external_insolation(sect,sat);
      if (sect->bodies[sat].orbradius>3e12) {
        temp=0;
      }
    }
    break;
  case BCLASS_BROWNDWARF:
    while (temp>4e4+randval_d(6e4)) {
      sect->bodies[sat].mass-=(sect->bodies[sat].mass/16);
      sect->bodies[sat].orbradius*=2;
      temp=external_insolation(sect,sat);
      if (sect->bodies[sat].orbradius>3e12) {
        temp=0;
      }
    }
    break;
  case BCLASS_GASGIANT:
    while (temp>2e4+randval_d(3e4)) {
      sect->bodies[sat].mass-=(sect->bodies[sat].mass/32);
      sect->bodies[sat].orbradius*=randval_i(2)+2;
      temp=external_insolation(sect,sat);
      if (sect->bodies[sat].orbradius>3e12) {
        temp=0;
      }
    }
    break;
  case BCLASS_TERRESTRIAL:
    while (temp>1e4+randval_d(4e3)) {
      sect->bodies[sat].mass/=2;
      sect->bodies[sat].orbradius*=2;
      temp=external_insolation(sect,sat);
      if (sect->bodies[sat].orbradius>3e12) {
        temp=0;
      }
    }
    break;
  case BCLASS_WHITEDWARF:
    while (temp>9e7) {
      sect->bodies[sat].mass/=1.4;
      sect->bodies[sat].orbradius*=5;
      temp=external_insolation(sect,sat);
      if (sect->bodies[sat].orbradius>3e12) {
        temp=0;
      }
    }
    sect->bodies[sat].class=BCLASS_NEUTRONSTAR;
    return;
    break;
  case BCLASS_NEUTRONSTAR:	/* These don't boil. */
  case BCLASS_BLACKHOLE:
  case BCLASS_WORMHOLE:
    break;
  }
  oldclass=sect->bodies[sat].class;
  initialize_body_class(&sect->bodies[sat]);
  initialize_body(&sect->bodies[sat]);
  if (sect->bodies[sat].class!=oldclass) {
    boil_orbiters(sect,sat);
  }
}

float grav_ratio(double x,double y, double d)
{
  /* Not sure of the right equation to use. */
  if (x>y) {
    return 1-(sqrt(y)/sqrt(x));
  } else {
    return sqrt(y)/sqrt(x);
  }
  //sqrt(sect->bodies[innerneighbour].mass)/sqrt(sect->bodies[body].mass);
}

void generate_orbiters(struct sector_t *sect, int body, int nesting)
{
  int	primary=sect->bodies[body].primary;
  double maxsatrange=2e11;
  int	innerneighbour=-1;
  int	outerneighbour=-1;
  float f;
  double aggregate_mass;
  double remaining_mass;
  double minrelmass=sect->bodies[body].mass/2e7;
  double minorb;
  int	numsats=0;
  int	sat;
  int	x;
  if (nesting>5) return; /* In case of runaway nesting loop. */
  if (primary>-1) {
    x=0;
    while (x<sect->numbodies) {
      if (sect->bodies[x].primary==primary) {
        if (sect->bodies[x].orbradius<sect->bodies[body].orbradius) {
          if (innerneighbour>=0) {
            if (sect->bodies[x].orbradius>sect->bodies[innerneighbour].orbradius) {
              innerneighbour=x;
            }
          } else {
            innerneighbour=x;
          }
        } else if (sect->bodies[x].orbradius>sect->bodies[body].orbradius) {
          if (outerneighbour>=0) {
            if (sect->bodies[x].orbradius<sect->bodies[outerneighbour].orbradius) {
              outerneighbour=x;
            }
          } else {
            outerneighbour=x;
          }
        } else if (x!=body) {
          /* return if a co-orbital satellite found - no moons will be
             allowed. */
          return;
        }
      }
      x++;
    }
  }
  /* Calculate orbital domains by mass ratios */
  if (primary>=0) {
    /* Orbital domain: sqrt(m1)/sqrt(m2) > oS/(oB/oS) */
//    f=sqrt(sect->bodies[primary].mass)/sqrt(sect->bodies[body].mass);
    f=grav_ratio(sect->bodies[body].mass,sect->bodies[primary].mass,
      sect->bodies[body].orbradius);
    maxsatrange=lesser_d(maxsatrange,sect->bodies[body].orbradius*f);
  }
  if (innerneighbour>=0) {
    f=grav_ratio(sect->bodies[body].mass,sect->bodies[innerneighbour].mass,
      sect->bodies[body].orbradius-sect->bodies[innerneighbour].orbradius);
    maxsatrange=lesser_d(maxsatrange, (sect->bodies[body].orbradius-
      sect->bodies[innerneighbour].orbradius)*f);
  }
  if (outerneighbour>=0) {
//    f=sqrt(sect->bodies[outerneighbour].mass)/sqrt(sect->bodies[body].mass);
    f=grav_ratio(sect->bodies[body].mass,sect->bodies[outerneighbour].mass,
      sect->bodies[outerneighbour].orbradius-sect->bodies[body].orbradius);
    maxsatrange=lesser_d(maxsatrange, (sect->bodies[outerneighbour].orbradius-
      sect->bodies[body].orbradius)*f);
  }
  aggregate_mass=sect->bodies[body].mass/(randval_d(2)+1.1);
  if (probability(0.5+((float)nesting/10))) 
    aggregate_mass /= randval_d(500)+500;
  remaining_mass=aggregate_mass;
  initialize_body(&sect->bodies[body]);
  minorb=(sect->bodies[body].radius*1.5)+1000;
  if ((minorb<1e6) && (maxsatrange>4e8)) {
    minorb+=1102123*randval_i(sect->bodies[body].mass/1e29);
  }
  while (remaining_mass>minrelmass) {
    extend_body_chain(sect,sect->numbodies);
    sat=sect->numbodies-1;
    switch (randval_i(6)) {
    case 0:
      /* Generate something Big.  Ioves/Saturns and the like. */
      sect->bodies[sat].mass=aggregate_mass/(randval_i(4)+1);
      break;
    case 1:
      /* Neptunes/Uranii */
      sect->bodies[sat].mass=randval_d(aggregate_mass)+minrelmass;
      break;
    default:
      sect->bodies[sat].mass=randval_d(minrelmass)+(minrelmass*randval_i(16));
    }
    sect->bodies[sat].primary=body;
    remaining_mass-=sect->bodies[sat].mass;
    if (sat>100) {
      remaining_mass-=(sect->bodies[sat].mass*sat);
    }
    initialize_body(&sect->bodies[sat]);
    sect->bodies[sat].orbradius=
      minorb+randval_d(minorb)+(sect->bodies[sat].radius*2);
    boil_orbiters(sect,sat);
    if (sect->bodies[sat].orbradius<=maxsatrange) {
      minorb=(sect->bodies[sat].radius*2)+randval_d(minorb)+
        sect->bodies[sat].orbradius+
        (randval_d(sect->bodies[sat].orbradius*4));
      if (probability(0.6-(0.3*nesting))) {
        generate_orbiters(sect,sat,nesting+1);
      }
      if (probability(numsats*0.02)) remaining_mass=0;
    } else {
/*      ui_debug("Boiled away...");*/
      sat--;
      sect->numbodies--;
      remaining_mass=0;
    }
  }
}

void generate_system(struct sector_t *sect)
{
  int b=sect->numbodies;
  extend_body_chain(sect,sect->numbodies);
  if (probability(0.08)) {
    sect->bodies[b].mass=6E28+(randval_d(1000)*1.0E29);
  } else if (probability(0.1)) {
    sect->bodies[b].mass=randval_d(10E30)+6E28;
  } else if (probability(0.4)) {
    sect->bodies[b].mass=randval_d(3.5E30)+6E28;
  } else if (probability(0.1)) {
    sect->bodies[b].mass=randval_d(1.6E29)+1e23;
  } else {
    sect->bodies[b].mass=randval_d(1.5E30)+6E28;
  }
/*  if (sect->bodies[b].mass<1.6E29) sect->bodies[b].mass+=1E29;*/
  initialize_body(&sect->bodies[b]);
  if (probability(0.5)) {
    generate_orbiters(sect,b,0);
  }
}

void generate_dwarf_system(struct sector_t *sect)
{
  ui_error("generate_dwarf_system() not yet written.");
}

int create_random_system(void)
{
  struct coord_t c;
  GString *s;
  struct sector_t sect;
  int fd;
  int retval=0;
  randomize_coords(&c);
  c.x+=(-5+randval_i(10));
  c.y+=(-5+randval_i(10));
  c.z+=(-5+randval_i(10));
  if (coords_oob(&c)) {
    return 0;
  }
  s=g_string_new("");
  g_string_sprintf(s,"%d/%d/%d",c.x,c.y,c.z);
  fd=create_sector(s->str);
  if (fd>=0) {
    read_sector(fd,&sect);
    if (sect.bodies==0) {
      if (probability(0.7)) {
        generate_system(&sect);
        if (probability(0.8)) {
          log_public("System generated",s->str);
        } else {
          log_private("System generated",s->str);
        }
        if (probability(0.2)) {
          generate_system(&sect);
          log_private("System generated",s->str);
          if (probability(0.4)) {
            generate_system(&sect);
          }
        }
        retval=1;
      }
    }
    write_sector(fd,&sect);
    dispose_sector(&sect);
    close_sector(fd);
  }
  g_string_free(s,TRUE);
  return retval;
}

char *body_name(struct body_t *b)
{
  if (b->name->len) {
    return b->name->str;
  }
  switch (b->class) {
  case BCLASS_STAR:		return "[Unnamed star]";
  case BCLASS_BROWNDWARF:	return "[Unnamed brown dwarf]";
  case BCLASS_GASGIANT:		return "[Unnamed gas giant]";
  case BCLASS_TERRESTRIAL:	return "[Unnamed terrestrial planet]";
  case BCLASS_WHITEDWARF:	return "[Unnamed white dwarf]";
  case BCLASS_NEUTRONSTAR:	return "[Unnamed neutron star]";
  case BCLASS_BLACKHOLE:	return "[Unnamed black hole]";
  case BCLASS_WORMHOLE:		return "[Unnamed wormhole]";
  }
  return "[Unnamed body]";
}

GString *body_mass_str(double mass)
{
  GString *s=g_string_new("");
  if (mass>1.6E29) {
    g_string_sprintf(s,"%f solar masses",(float)(mass/1.99E30));
  } else if (mass>1.900e26) {
    g_string_sprintf(s,"%f jupiter masses",(float)(mass/1.900e27));
  } else if (mass>5.972e18) {
    g_string_sprintf(s,"%f earth masses",(float)(mass/5.972e24));
  } else {
    g_string_sprintf(s,"%f kilograms",(float)mass);
  }
  return s;
}

const char *body_class_string(int class)
{
  switch (class) {
  case BCLASS_STAR:		return "Star";
  case BCLASS_BROWNDWARF:	return "Brown Dwarf";
  case BCLASS_GASGIANT:	  	return "Gas Giant";
  case BCLASS_TERRESTRIAL:	return "Terrestrial planet";
  case BCLASS_WHITEDWARF:	return "White Dwarf";
  case BCLASS_NEUTRONSTAR:	return "Neutron Star";
  case BCLASS_BLACKHOLE:	return "Black Hole";
  case BCLASS_WORMHOLE:		return "Wormhole";
  }
  return "Unknown body type.";
}
